// $Modified: Tue Ju  6 09:31:07 20: Thu Jul  8 09:30:45 2010 by uwer 10 by uwer $
#include "Hathor.h"
#include "HathorFits.h"
#include "HathorPdf.h"
#include "HathorVegas.h"

#define PR(_X_) cout << "HATHOR: " << #_X_ << " " << _X_ << endl;


using namespace std;

const double Hathor::z2  = 1.6449340668482264365;   /* Zeta(2) */
const double Hathor::z3  = 1.2020569031595942854;   /* Zeta(3) */
const double Hathor::ln2 = 0.69314718055994530942; /* log(2)  */

const double Hathor::inv4pi  = 1./(4.*M_PI);
const double Hathor::inv4pi2 = inv4pi*inv4pi;
const double Hathor::inv4pi4 = inv4pi2*inv4pi2;

const double Hathor::ca  = 3.0;
const double Hathor::ca2 = ca*ca;
const double Hathor::cf  = 4.0/3.0;

double Hathor::hcq = 0.389379323e9;

void Hathor::info(std::string s){
  if (debug){
    std::cout << " HATHOR: " << s << std::endl;
  }
}

void Hathor::info(std::string s, double x){
  if (debug){
    std::cout << " HATHOR: " << s << " " << x << std::endl;
  }
}

void Hathor::warn(std::string s){
    std::cerr << " HATHOR: WARNING:" << s << std::endl;
}

void Hathor::err(std::string s){
    std::cerr << " HATHOR: ERROR:" << s << std::endl;
}

void Hathor::sethc2(double hcq_){
  hcq = hcq_;
  info("(hc)^2 set to", hcq);
}

void Hathor::PrintOptions(){
  cout << " HATHOR: ";
  if ( LO & scheme )
    cout << "Hathor::LO ";
  if ( NLO & scheme )
    cout << "Hathor::NLO ";
  if ( NNLO & scheme )
    cout << "Hathor::NNLO ";
  if ( LOG_ONLY & scheme )
    cout << "Hathor::LOG_ONLY ";
  if ( PDF_SCAN & scheme )
    cout << "Hathor::PDF_SCAN ";
  if ( PDF_SYM_ERR & scheme )
    cout << "Hathor::PDF_SYM_ERR ";  
  if ( MS_MASS & scheme )
    cout << "Hathor::MS_MASS ";
  cout << endl;
}

void Hathor::setColliderType(COLLIDERTYPE type){
  
  if (PP == type){
      shad = 7000.*7000;
      collider = PP;
      info("Collider set to Proton-Proton");
      info("sqrt(shad) is set to [GeV]: ", sqrt(shad));
      info("Change if needed with setSqrtShad(newval)");
      return;
  }

  if (type == PPBAR){
    shad = 1960.*1960;
    collider = PPBAR;
    info("Collider set to Proton-AntiProton");
    info("sqrt(shad) is set to [GeV]: ", sqrt(shad));
    info("Change if needed with setSqrtShad(newval)");
    return;
  }
  // By construction we should never arrive here;
  err("Unknown collider used in setColliderType");
  exit(1);
}

void Hathor::setSqrtShad(double ecms){
  shad = ecms*ecms;
  info("Collider energy set to [GeV]: ",ecms);
}

void Hathor::setNf(int n){
  nf = n;
  info("Number of massless flavours set to", nf);
  info("Note that this will not affect the scheme for alpha_s");
  info("provided by the pdf set.");
  setnf_(nf);
}

void Hathor::setScheme(unsigned int newscheme){

  if ( (LOG_ONLY & newscheme) ){
    if ( MS_MASS & newscheme ){
      err("Incompatible use: In LOG_ONLY the MS_MASS scheme");
      err("is not supported. If your really believe you need it,");
      err("please contact the authors.");
      exit(1);
    }
    if ( ! (NNLO & newscheme) ){
      err("LOG_ONLY make only sense in NNLO order");
      err("Please use in addition Hathor::NNLO as option");
    }
  }

  if ( PDF_SCAN & newscheme ){
    pdfmax = pdf.NumberPdf()+1;
    if (pdfmax<=2){
      err("Pdf does not seem to support error pdfs, at least ");
      err("the number of sets returned by the pdf is <= 1");
      exit(1);
    }
  }
  else
    pdfmax = 1;
  scheme = newscheme;
}

void Hathor::setCqq(double tmp){
  Cqq = tmp;
  info("Cqq set to ", Cqq);
};

void Hathor::setCgg(double tmp){
  Cgg = tmp;
  info("Cgg set to ", Cgg);
};

void Hathor::setPrecision(int n){
  calls = n;
  info("Precision set to ", calls);
};

void Hathor::getResult(int pdfset, double & integral_, 
		       double & err_, double chi2a_){

  if ( ! (scheme & PDF_SCAN ) && (pdfset>0) ){
    err("Use the option PDF_SCAN to evaluate pdf uncertainties");
    exit(1);
  }

  integral_ = integral[pdfset]; 
  err_ = error[pdfset];
  chi2a_ = chi2a[pdfset];
  
  if ( fabs(err_/integral_ ) > 0.02 ){
    warn("Numerical integration error larger than 2 %");
    warn("Consider reruning with setPrecision set to a higher value");
  }

  if ( fabs(err_/integral_ ) > 0.1 ){
    err("Numerical integration returned error larger 10 %");
    err("Result is unreliable rerun with setPrecision set to higher value");
  }

}

void Hathor::getPdfErr(double & up, double & down){

  if ( ! ( PDF_SCAN & scheme ) ){
    err("Use the PDF_SCAN to evaluate pdf uncertainties");
    exit(1);
  }

  up=0; down=0;

  if ( PDF_SYM_ERR & scheme ) {
    for(int i=1; i < pdfmax; i++){
      up += pow(integral[i]-integral[0],2);
    }
    up = sqrt(up); 
    down = sqrt(down);
    return;
  }

  double tmp;
  for(int i=0; i < pdfmax/2; i++){
    tmp = max(integral[2*i+1]-integral[0],integral[2*i+2]-integral[0]); 
    tmp = max(tmp,0.);
    up += tmp*tmp;
    tmp = max(integral[0]-integral[2*i+1],integral[0]-integral[2*i+2]); 
    tmp = max(tmp,0.);
    down += tmp*tmp;
  }
  up = sqrt(up); down = sqrt(down);
}

Hathor::Hathor(Pdf & pdf_) : collider(PP), pdfmax(1), nf(5), calls(MEDIUM),
			     shad(14000.*14000), pdf(pdf_)  {
  debug = 1;
  Cqq = 0;
  Cgg = 0;
  scheme = LO | NLO | NNLO | POLE_MASS;  
  update();
}

void Hathor::update(){
  /*
   * Update constants in case that the setting has changed,
   * i.e. alphas value, nf, or the scales, etc...
   */
  as = pdf.GetAlphas(mur);
  as2 = as*as;
  a = as/M_PI;
  as_pdf[0] = as;

  if ( PDF_SCAN & scheme ){
    for (int ipdf = 1; ipdf < pdfmax; ipdf++){
      pdf.InitMember(ipdf);
      as_pdf[ipdf] = pdf.GetAlphas(mur);
    } 
  }
  m2 = m*m;
  muf2 = muf*muf;
  mur2 = mur*mur;
  Lmf = log(muf2/m2);
  Lmr = log(mur2/muf2);
  


  /* Update QCD beta function in case that nf has changed */
  b0 = ( 11.0/3.0*ca - 4.0/3.0*nf/2.0 ) * inv4pi2;
  b1 = (+ 34.0/3.0*ca2 - 20.0/3.0*ca*nf/2.0 - 4.0*cf*nf/2.0 ) * inv4pi4;
  
  /* 
   * In the definition of d1,d2 the renormalisation scale mur 
   * is set to m, as a consequence no logs appear here.
   */
  d1 = 4./3.;
  d2 = 307./32. + 2.*z2 + 2./3.*z2*ln2 - 1./6.*z3
    - ( 71./144. + 1./3.*z2 )*nf;
  /*
   * Update nf in the Fortran fit functions:
   */
  setnf_(nf);
  /*
   * Clear result from last integration
   */
  for(int i=0; i<100;i++){
    integral[i] = 0; error[i]=0; chi2a[i]=0;
  }
}

void Hathor::evaluateNLO(double beta){
  fgg10 = tfgg10_(beta); fgg11 = tfgg11_(beta);  
  fqq10 = tfqq10_(beta); fqq11 = tfqq11_(beta);
  fgq10 = tfgq10_(beta); fgq11 = tfgq11_(beta);

  nlogg = 4.0 * M_PI * as * (fgg10 + Lmf * fgg11 + 2.0*b0 * Lmr * fgg00);
  nloqq = 4.0 * M_PI * as * (fqq10 + Lmf * fqq11 + 2.0*b0 * Lmr * fqq00);
  nlogq = 4.0 * M_PI * as * (fgq10 + Lmf*fgq11);
}

void Hathor::evaluateNNLO(double beta){
  
  double xfqq00, xfgg00;
  double not_log_enhanced = 1.;

  if ( LOG_ONLY & scheme ) {
    
    double lnbeta  = log(beta);
    double lnbeta2 = lnbeta*lnbeta;
    double lnbeta3 = lnbeta2*lnbeta;
    double lnbeta4 = lnbeta3*lnbeta;

    not_log_enhanced = 0.;
    //xfqq00 = fqq00; xfgg00 = fgg00;
    xfqq00 = M_PI/9.*beta; xfgg00 = 7./192.*M_PI*beta;

    fqq20 = xfqq00 * inv4pi4 * 
      ( + 3.60774/beta/beta
	+ 1/beta*(-140.368*lnbeta2+32.106*lnbeta+3.95105)
	+ 910.222*lnbeta4-1315.53*lnbeta3+592.292*lnbeta2
	+ 528.557*lnbeta);
    
    fgg20 = xfgg00 * inv4pi4 * 
      ( + 68.5471/beta/beta
	+ 1/beta*(496.3*lnbeta2+321.137*lnbeta-8.62261)
	+ 4608*lnbeta4 - 1894.91*lnbeta3 
	- 912.349*lnbeta2+2456.74*lnbeta);
    
    fqq21 = xfqq00 * inv4pi4 *
      ( + 1/beta*(-49.3480 + 70.1839*lnbeta)
	- 910.222*lnbeta3+1358.99*lnbeta2
	- 476.012*lnbeta+163.524);
    
    fgg21 = xfgg00 * inv4pi4 * 
      ( + 1/beta*(39.6351-248.150*lnbeta)
	-4608*lnbeta3+2606.77*lnbeta2+368.035*lnbeta
	-171.000);
    
    fqq22 = xfqq00 * inv4pi4 * (227.555*lnbeta2-377.874*lnbeta+96.3456);
    
    fgg22 = xfgg00 * inv4pi4 * (1152*lnbeta2-890.989*lnbeta-104.291);
    
    xfqq10 = xfqq00 * inv4pi2 * (42.66667*lnbeta2
			       -20.6105*lnbeta+14.5312-3.28986/beta);
    xfgg10 = xfgg00 * inv4pi2 * (96*lnbeta2
			       -9.51647*lnbeta+25.8548+5.16979/beta);
    
    xfqq11 = xfqq00 * inv4pi2 * (-21.3333*lnbeta+13.8795);
    xfgg11 = xfgg00 * inv4pi2 * (-48*lnbeta+14.7289);

    nnlogq = 0;

  } else {
    fgg20 = tfgg20_(beta); fgg21 = tfgg21_(beta); fgg22 = tfgg22_(beta);
    fqq20 = tfqq20_(beta); fqq21 = tfqq21_(beta); fqq22 = tfqq22_(beta);
    xfqq10 = fqq10; xfgg10 = fgg10; xfqq11 = fqq11; xfgg11 = fgg11;
    xfqq00 = fqq00; xfgg00 = fgg00;

    nnlogq = pow(4.0 * M_PI * as,2) * 
      (
       + fgq20 + Lmf*fgq21 + Lmf*Lmf*fgq22
       + 3.0*b0*Lmr* fgq10 
       + 3.0*b0*Lmr*Lmf* fgq11
       );  
  }    

  fgq20 = tfgq20_(beta); fgq21 = tfgq21_(beta); fgq22 = tfgq22_(beta);
  
  nnlogg = pow(4.0 * M_PI * as,2) * 
    ( + fgg20 + Lmf*fgg21 + Lmf*Lmf*fgg22 
      + Lmr * ( 3.0*b0*xfgg10 + 2.0 * b1 * fgg00 * not_log_enhanced )
      + 3.0 * b0 * Lmr*Lmf* xfgg11  
      + 3.0 * b0*b0 * Lmr*Lmr* fgg00 * not_log_enhanced 
      + xfgg00*Cgg*inv4pi4 );  
  
  nnloqq = pow(4.0 * M_PI * as,2) * 
    ( + fqq20 + Lmf*fqq21 + Lmf*Lmf*fqq22 
      + Lmr * ( 3.0 * b0 * xfqq10 + 2.0 * b1 * fqq00 * not_log_enhanced )
      + 3.0 * b0 * Lmr*Lmf * xfqq11 
      + 3.0 * b0*b0 * Lmr*Lmr * fqq00 * not_log_enhanced 
      + xfqq00*Cqq*inv4pi4 );  
    
}

double Hathor::f(const double x[2],const double wgt, double res[]){
  
  /*
   * Differential cross section including the parton luminosities
   */
  double sparton = x[0]*(shad-4*m2)+4*m2;
  double q = x[1]*(shad-sparton)+sparton;

  double jacobian1 = (shad-4*m2)*(shad-sparton)/shad/q;

  double x1=q/shad, x2=sparton/q;

  double delta = x2*0.00001;

  if ( x2+delta > 1. )
    return 0.;

  /*
   * Kinematical variables
   */
  double beta = sqrt(1.- 4.*m2/sparton);
  double rho =   4.*m2/sparton;
  double beta2 = 1.- rho;
  double beta4 = beta2*beta2;


  /* leading-order scaling functions */
  fgg00 = tfgg00_(beta);
  fqq00 = tfqq00_(beta);

  /* next-to-leading-order scaling functions and cross sections*/ 
  if ( ( scheme & NNLO ) || ( scheme & NLO ) )
    evaluateNLO(beta);
  
  /* next-to-next-to-leading order scaling functions and cross sections: */  
  if ( scheme & NNLO )
    evaluateNNLO(beta); 

  double hi[13],hj[13],hjleft[13],hjright[13];

  for(int ipdf=0; ipdf<pdfmax; ipdf++){

    double lo=0,nlo=0,nnlo=0;

    pdf.InitMember(ipdf);
    pdf.GetPdf(x1,muf,hi);
    pdf.GetPdf(x2,muf,hj);
    pdf.GetPdf(x2-delta,muf,hjleft);
    pdf.GetPdf(x2+delta,muf,hjright);

    if ( PPBAR == collider ){
      ChargeConjugation(hj);
      ChargeConjugation(hjleft);
      ChargeConjugation(hjright);
    }

    /*
     * Partonfluxes and their derivatives:
     */    
    double gg  = fluxgg(hi,hj);
    double qqb = fluxqqb(hi,hj);
    double qg  = fluxqg(hi,hj);
    double gqb = fluxgqb(hi,hj);
    
    double dgg =  x2/(2.*delta) * ( fluxgg(hi,hjright) -  fluxgg(hi,hjleft) );
    double dqqb = x2/(2.*delta) * (fluxqqb(hi,hjright) - fluxqqb(hi,hjleft) );
    double dqg =  x2/(2.*delta) * ( fluxqg(hi,hjright) -  fluxqg(hi,hjleft) );
    double dgqb = x2/(2.*delta) * (fluxgqb(hi,hjright) - fluxgqb(hi,hjleft) );
    
    if ( scheme & LO ) {
      lo =  gg * fgg00 + qqb * fqq00 ;
    }

    if ( scheme & NLO ) {
      nlo =  gg  * nlogg + qqb * nloqq + ( qg + gqb ) * nlogq ;
      if ( scheme & MS_MASS ){
	nlo += + 2.* d1 * a * ( dgg * fgg00 + dqqb * fqq00 );
      }
    }
    
    if ( scheme & NNLO ){
      nnlo = gg * nnlogg + qqb * nnloqq + ( qg + gqb ) * nnlogq ;
      
      if ( scheme & MS_MASS ) {
	double mdiffgg = +1./192.*M_PI*rho/beta 
	  * ( beta*(36.-40*beta2+4*beta4)*log((1.+beta)/(1.-beta))
	      -7.-116.*beta2+91.*beta4);
	double mdiffqq = -1./9.*M_PI*pow(rho,3)/beta;
	
	
	nnlo += 
	  + dgg * a * ( + (2.*d2-d1*d1) * a * fgg00 + 2.*d1 * nlogg
			+ d1*d1 * a *mdiffgg ) 
	  + dqqb * a * ( + (2.*d2-d1*d1) * a * fqq00 + 2.*d1 * nloqq
			+ d1*d1 * a * mdiffqq ) 
	  + ( dqg + dgqb ) * 2.*d1 * a * nlogq
	  + (-2.*d1*a*4.*M_PI*as) * ( gg*fgg11 + qqb*fqq11 + (qg+gqb)*fgq11 )
	  + 2.*d1* a * 4.*M_PI*as* log(mur2/m2)*b0 
	  * ( dgg * fgg00 + dqqb * fqq00) ;
      }
    }
    
    //result *= as2/m2  * jacobian1 * hcq;
    double asi = as_pdf[ipdf];
    res[ipdf] = as2 * pow(asi/as,2) / m2  * jacobian1 * hcq 
      * ( lo + (asi/as) * nlo + pow(asi/as,2) * nnlo );
  }
  
  return( res[0] );
  
}

int Hathor::getPdfNumber(){ return(pdfmax);};

double Hathor::getXsection(double m_, double mur_,double muf_){
  m = m_; mur=mur_; muf=muf_;
  update();
  
  Vegas<Hathor> integrator(*this,pdfmax);
  //  integrator.setnprn(1);
  integrator.vegas();

  for (int i=0; i<pdfmax; i++){
    integral[i] = integrator.avgi_[i];
    error[i] = integrator.sd_[i];
    chi2a[i] = integrator.chi2a_[i];
  }

  return(integrator.avgi);

}

